using NUnit.Framework;
using FluentAssertions;

using Encompass;

using System;
using System.Linq;
using System.Collections.Generic;
using Encompass.Exceptions;

namespace Tests
{
    public class EngineTest
    {
        static List<KeyValuePair<Guid, MockComponent>> resultComponents;
        static MockComponent resultComponent;

        static List<MockMessage> resultMessages;

        public class ReadComponentsTestEngine : Engine
        {
            public override void Update(double dt)
            {
                resultComponents = this.ReadComponents<MockComponent>().ToList();
            }
        }

        public class ReadComponentTestEngine : Engine
        {
            public override void Update(double dt)
            {
                resultComponent = this.ReadComponent<MockComponent>().Value;
            }
        }

        [Test]
        public void ReadComponents()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new ReadComponentsTestEngine());

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            MockComponent mockComponentB;
            mockComponentB.myInt = 1;
            mockComponentB.myString = "howdy";

            var componentAID = entity.AddComponent(mockComponent);
            var componentBID = entity.AddComponent(mockComponentB);

            var world = worldBuilder.Build();

            world.Update(0.01f);

            var resultComponentValues = resultComponents.Select((kv) => kv.Value);
            resultComponentValues.Should().Contain(mockComponent);
            resultComponentValues.Should().Contain(mockComponentB);
        }

        [Test]
        public void ReadComponent()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new ReadComponentTestEngine());

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            entity.AddComponent(mockComponent);

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.AreEqual(mockComponent, resultComponent);
        }

        [Test]
        public void ReadComponentWhenMultipleComponents()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new ReadComponentTestEngine());

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            MockComponent mockComponentB;
            mockComponentB.myInt = 1;
            mockComponentB.myString = "howdy";

            entity.AddComponent(mockComponent);
            entity.AddComponent(mockComponentB);

            var world = worldBuilder.Build();

            Assert.Throws<InvalidOperationException>(() => world.Update(0.01f));
        }

        [Mutates(typeof(MockComponent))]
        public class UpdateComponentTestEngine : Engine
        {
            public override void Update(double dt)
            {
                (var componentID, var component) = this.ReadComponent<MockComponent>();

                component.myInt = 420;
                component.myString = "blaze it";
                this.UpdateComponent(componentID, component);

                resultComponent = this.ReadComponent<MockComponent>().Value;
            }
        }

        [Test]
        public void UpdateComponent()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new UpdateComponentTestEngine());

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            entity.AddComponent(mockComponent);

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.AreEqual(420, resultComponent.myInt);
            Assert.AreEqual("blaze it", resultComponent.myString);
        }

        public class UndeclaredUpdateComponentTestEngine : Engine
        {
            public override void Update(double dt)
            {
                (var componentID, var component) = this.ReadComponent<MockComponent>();

                component.myInt = 420;
                component.myString = "blaze it";
                this.UpdateComponent(componentID, component);

                component = this.ReadComponent<MockComponent>().Value;
            }
        }

        [Test]
        public void UpdateUndeclaredComponent()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new UndeclaredUpdateComponentTestEngine());

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            entity.AddComponent(mockComponent);

            var world = worldBuilder.Build();

            var ex = Assert.Throws<IllegalComponentMutationException>(() => world.Update(0.01f));
            Assert.That(ex.Message, Is.EqualTo("Engine UndeclaredUpdateComponentTestEngine tried to mutate undeclared Component MockComponent"));
        }

        struct MockMessage : IMessage
        {
            public string myString;
        }

        [Emits(typeof(MockMessage))]
        public class MessageEmitEngine : Engine
        {
            public override void Update(double dt)
            {
                MockMessage message;
                message.myString = "howdy";

                this.EmitMessage(message);
            }
        }

        [Reads(typeof(MockMessage))]
        public class MessageReadEngine : Engine
        {
            public override void Update(double dt)
            {
                resultMessages = this.ReadMessages<MockMessage>().ToList();
            }
        }

        [Test]
        public void EmitAndReadMessage()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new MessageEmitEngine());
            worldBuilder.AddEngine(new MessageReadEngine());

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.AreEqual(resultMessages.First().myString, "howdy");
        }

        public class UndeclaredMessageEmitEngine : Engine
        {
            public override void Update(double dt)
            {
                MockMessage message;
                message.myString = "howdy";

                this.EmitMessage(message);
            }
        }

        static IEnumerable<MockMessage> emptyReadMessagesResult;
        [Reads(typeof(MockMessage))]
        class ReadMessagesWhenNoneExistEngine : Engine
        {
            public override void Update(double dt)
            {
                emptyReadMessagesResult = ReadMessages<MockMessage>();
            }
        }

        [Test]
        public void ReadMessagesWhenNoneHaveBeenEmitted()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new ReadMessagesWhenNoneExistEngine());

            var world = worldBuilder.Build();

            world.Update(0.01f);

            emptyReadMessagesResult.Should().BeEmpty();
        }

        [Test]
        public void EmitUndeclaredMessage()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new UndeclaredMessageEmitEngine());

            var world = worldBuilder.Build();

            var ex = Assert.Throws<IllegalMessageEmitException>(() => world.Update(0.01f));
            Assert.That(ex.Message, Is.EqualTo("Engine UndeclaredMessageEmitEngine tried to emit undeclared Message MockMessage"));
        }

        static bool someTest;

        [Emits(typeof(MockMessage))]
        class EmitMockMessageEngine : Engine
        {
            public override void Update(double dt)
            {
                MockMessage message;
                message.myString = "howdy";

                this.EmitMessage(message);
            }
        }

        [Reads(typeof(MockMessage))]
        class SomeTestEngine : Engine
        {
            public override void Update(double dt)
            {
                someTest = this.Some<MockMessage>();
            }
        }

        [Test]
        public void Some()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new EmitMockMessageEngine());
            worldBuilder.AddEngine(new SomeTestEngine());

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.That(someTest, Is.True);
        }

        class UndeclaredSomeEngine : Engine
        {
            public override void Update(double dt)
            {
                someTest = this.Some<MockMessage>();
            }
        }

        [Test]
        public void IllegalSome()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new EmitMockMessageEngine());
            worldBuilder.AddEngine(new UndeclaredSomeEngine());

            var world = worldBuilder.Build();

            Assert.Throws<IllegalMessageReadException>(() => world.Update(0.01f));
        }

        static KeyValuePair<Guid, MockComponent> pairA;
        static KeyValuePair<Guid, MockComponent> pairB;

        class SameValueComponentReadEngine : Engine
        {
            public override void Update(double dt)
            {
                var components = ReadComponents<MockComponent>();

                pairA = components.First();
                pairB = components.Last();
            }
        }

        // Tests that components with identical values should be distinguishable by ID
        [Test]
        public void SameValueComponents()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new SameValueComponentReadEngine());

            MockComponent componentA;
            componentA.myInt = 20;
            componentA.myString = "hello";

            MockComponent componentB;
            componentB.myInt = 20;
            componentB.myString = "hello";

            var entity = worldBuilder.CreateEntity();
            entity.AddComponent(componentA);
            entity.AddComponent(componentB);

            var world = worldBuilder.Build();
            world.Update(0.01f);

            Assert.That(pairA, Is.Not.EqualTo(pairB));
            Assert.That(pairA.Value, Is.EqualTo(pairB.Value));
        }

        static IEnumerable<KeyValuePair<Guid, MockComponent>> emptyComponentReadResult;

        class ReadEmptyMockComponentsEngine : Engine
        {
            public override void Update(double dt)
            {
                emptyComponentReadResult = ReadComponents<MockComponent>();
            }
        }

        [Test]
        public void ReadComponentsOfTypeWhereNoneExist()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new ReadEmptyMockComponentsEngine());

            var world = worldBuilder.Build();
            world.Update(0.01f);

            Assert.That(emptyComponentReadResult, Is.Empty);
        }

        struct DestroyerComponent : IComponent { }

        class DestroyerEngine : Engine
        {
            public override void Update(double dt)
            {
                var componentPairs = ReadComponents<DestroyerComponent>();

                foreach (var componentPair in componentPairs)
                {
                    var componentID = componentPair.Key;
                    var entityID = GetEntityIDByComponentID(componentID);
                    Destroy(entityID);
                }
            }
        }

        static IEnumerable<KeyValuePair<Guid, MockComponent>> results;
        class ReaderEngine : Engine
        {
            public override void Update(double dt)
            {
                results = ReadComponents<MockComponent>();
            }
        }

        [Test]
        public void DestroyEntity()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new DestroyerEngine());
            worldBuilder.AddEngine(new ReaderEngine());

            var entity = worldBuilder.CreateEntity();
            var entityB = worldBuilder.CreateEntity();

            DestroyerComponent destroyerComponent;
            MockComponent mockComponent;
            mockComponent.myInt = 2;
            mockComponent.myString = "blah";

            entity.AddComponent(destroyerComponent);
            var componentID = entity.AddComponent(mockComponent);

            entityB.AddComponent(destroyerComponent);
            var componentBID = entityB.AddComponent(mockComponent);

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.That(results, Does.Not.Contain(new KeyValuePair<Guid, MockComponent>(componentID, mockComponent)));
            Assert.That(results, Does.Not.Contain(new KeyValuePair<Guid, MockComponent>(componentBID, mockComponent)));
        }

        static Entity entityFromComponentIDResult;
        class GetEntityFromComponentIDEngine : Engine
        {
            public override void Update(double dt)
            {
                var componentID = ReadComponent<MockComponent>().Key;
                entityFromComponentIDResult = GetEntityByComponentID(componentID);
            }
        }

        [Test]
        public void GetEntityFromComponentID()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new GetEntityFromComponentIDEngine());

            MockComponent component;
            component.myInt = 2;
            component.myString = "howdy";

            var entity = worldBuilder.CreateEntity();
            entity.AddComponent(component);

            var world = worldBuilder.Build();
            world.Update(0.01f);

            Assert.That(entity, Is.EqualTo(entityFromComponentIDResult));
        }

        static MockComponent mockComponentByIDResult;
        class GetComponentByIDEngine : Engine
        {
            public override void Update(double dt)
            {
                var componentID = ReadComponent<MockComponent>().Key;
                mockComponentByIDResult = GetComponentByID<MockComponent>(componentID);
            }
        }
        [Test]
        public void GetComponentByID()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new GetComponentByIDEngine());

            MockComponent component;
            component.myInt = 2;
            component.myString = "howdy";

            var entity = worldBuilder.CreateEntity();
            entity.AddComponent(component);

            var world = worldBuilder.Build();
            world.Update(0.01f);

            Assert.That(component, Is.EqualTo(mockComponentByIDResult));
        }

        struct OtherComponent : IComponent { }

        class GetComponentByIDWithTypeMismatchEngine : Engine
        {
            public override void Update(double dt)
            {
                var componentID = ReadComponent<MockComponent>().Key;
                GetComponentByID<OtherComponent>(componentID);
            }
        }

        [Test]
        public void GetComponentByIDWithTypeMismatch()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine(new GetComponentByIDWithTypeMismatchEngine());

            MockComponent component;
            component.myInt = 2;
            component.myString = "howdy";

            var entity = worldBuilder.CreateEntity();
            entity.AddComponent(component);

            var world = worldBuilder.Build();

            Assert.Throws<ComponentTypeMismatchException>(() => world.Update(0.01f));
        }
    }
}
